//----------------------------------------------------------------------------
//  Sound Blitter
//----------------------------------------------------------------------------
// 
//  Copyright (c) 2010-2012 Andrew Apted
//  Copyright (c) 1999-2009 The EDGE Team.
//  Copyright (C) 1993-1996 by id Software, Inc.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code by Id Software.
//
//----------------------------------------------------------------------------

#include "main.h"

#include "snd_blit.h"


// Sound must be clipped to prevent distortion (clipping is
// a kind of distortion of course, but it's much better than
// the "white noise" you get when values overflow).
//
// The more safe bits there are, the less likely the final
// output sum will overflow into white noise, but the less
// precision you have left for the volume multiplier.
#define SAFE_BITS  4
#define CLIP_THRESHHOLD  ((1L << (31-SAFE_BITS)) - 1)

// This value is how much breathing room there is until
// sounds start clipping.  More bits means less chance
// of clipping, but every extra bit makes the sound half
// as loud as before.
int var_quiet_factor = 1;  // FIXME: CVAR


#define QUEUE_SUPPORT  1


mix_channel_t  mix_chan[NUM_CHANNELS];

static int * mix_buffer;
static int   mix_buf_len;



int sfx_volume = 0;

int var_swap_stereo = 0;  // FIXME: CVAR

// these are analogous to viewx/y/z/angle
float listen_x;
float listen_y;
float listen_z;
float listen_angle;  // FIXME: REMOVE

vec3_t  listen_forward;
vec3_t  listen_right;
vec3_t  listen_up;


// this is incremented every time a new sound is played
static int sfx_start_count;


#ifdef QUEUE_SUPPORT

typedef struct queue_sound_s
{
	bool_t  playing;

	sound_t * sfx;

} queue_sound_t;

// this equals about 5 seconds of 44Khz stereo sound
#define MAX_QUEUE_BUFS     1024
#define QUEUE_BUF_SAMPLES  512    // or 256 pairs

static volatile queue_sound_t  q_bufs[MAX_QUEUE_BUFS];

static int free_q_index;   // <-- not touched by mixing thread

static int playing_q_index;  // <--- not modified by normal thread


static mix_channel_t  music_chan;

#endif


#if 0
// -AJA- 2005/02/26: table to convert slider position to GAIN.
//       Curve was hand-crafted to give useful distinctions of
//       volume levels at the quiet end.  Entry zero always
//       means total silence (in the table for completeness).
float slider_to_gain[SND_SLIDER_NUM] =
{
	0.00000, 0.00200, 0.00400, 0.00800, 0.01600,
	0.03196, 0.05620, 0.08886, 0.12894, 0.17584,
	0.22855, 0.28459, 0.34761, 0.41788, 0.49553,
	0.58075, 0.67369, 0.77451, 0.88329, 1.00000
};
#endif


static void ComputeDelta(mix_channel_t * chan, sound_t * sfx)
{
	// frequency close enough ?
	if (sfx->rate > (snd.rate - snd.rate/100) &&
		sfx->rate < (snd.rate + snd.rate/100))
	{
		chan->delta = (1 << 10);
	}
	else
	{
		chan->delta = (fixed22_t) floor(sfx->rate * 1024.0f / snd.rate);
	}
}


static void ComputeVolume(mix_channel_t * chan)
{
	float MAX_VOL = (1 << (16 - SAFE_BITS - (var_quiet_factor-1))) - 3;


	// FIXME
	{
		chan->volume_L = MAX_VOL / 2.0;
		chan->volume_R = MAX_VOL / 2.0;

		return;
	}


	float sep = 0.5f;
	float mul = 1.0f;

	if (0)  // FIXME??  && category >= SNCAT_Opponent)
	{
#if 0  // FIXME
		if (snd.channels >= 2)
		{
			angle_t angle = R_PointToAngle(listen_x, listen_y, pos->x, pos->y);

			// same equation from original DOOM 
			sep = 0.5f - 0.38f * M_Sin(angle - listen_angle);
		}

		if (! boss)
		{
			float dist = P_ApproxDistance(listen_x - pos->x, listen_y - pos->y, listen_z - pos->z);

			// -AJA- this equation was chosen to mimic the DOOM falloff
			//       function, but instead of cutting out @ dist=1600 it
			//       tapers off exponentially.
			mul = exp(-MAX(1.0f, dist - S_CLOSE_DIST) / 800.0f);
		}
#endif
	}

// FIXME	MAX_VOL = MAX_VOL * mul * volume.value;

	// strictly linear equations
	chan->volume_L = (int) (MAX_VOL * (1.0 - sep));
	chan->volume_R = (int) (MAX_VOL * (0.0 + sep));

	if (var_swap_stereo)  /* SWAP ! */
	{
		int tmp        = chan->volume_L;
		chan->volume_L = chan->volume_R;
		chan->volume_R = tmp;
	}
}


static void ComputeMusicVolume(mix_channel_t * chan)
{
	// the MAX_VOL value here is equivalent to the 'NORMAL' quiet
	// factor.  If the music uses the full range, then the output
	// will also use the full range without being clipped
	// (assuming no other sounds are playing).
	//
	// We do not use 'quiet_factor' here because that mainly applies
	// to the mixing of game/UI sounds.  The music volume slider allows
	// for quieter output.

	float MAX_VOL = (1 << (16 - SAFE_BITS)) - 3;

 	MAX_VOL = MAX_VOL * mus_volume.value;

	chan->volume_L = (int) MAX_VOL;
	chan->volume_R = (int) MAX_VOL;
}


//----------------------------------------------------------------------------
//  LOW LEVEL MIXING CODE
//----------------------------------------------------------------------------

static void BlitToU8(const int *src, u8_t *dest, int length)
{
	const int *s_end = src + length;

	while (src < s_end)
	{
		int val = *src++;

		     if (val >  CLIP_THRESHHOLD) val =  CLIP_THRESHHOLD;
		else if (val < -CLIP_THRESHHOLD) val = -CLIP_THRESHHOLD;

		*dest++ = (u8_t) ((val >> (24-SAFE_BITS)) ^ 0x80);
	}
}

static void BlitToS8(const int *src, s8_t *dest, int length)
{
	const int *s_end = src + length;

	while (src < s_end)
	{
		int val = *src++;

		     if (val >  CLIP_THRESHHOLD) val =  CLIP_THRESHHOLD;
		else if (val < -CLIP_THRESHHOLD) val = -CLIP_THRESHHOLD;

		*dest++ = (s8_t) (val >> (24-SAFE_BITS));
	}
}

static void BlitToU16(const int *src, u16_t *dest, int length)
{
	const int *s_end = src + length;

	while (src < s_end)
	{
		int val = *src++;

		     if (val >  CLIP_THRESHHOLD) val =  CLIP_THRESHHOLD;
		else if (val < -CLIP_THRESHHOLD) val = -CLIP_THRESHHOLD;

		*dest++ = (u16_t) ((val >> (16-SAFE_BITS)) ^ 0x8000);
	}
}

static void BlitToS16(const int *src, s16_t *dest, int length)
{
	const int *s_end = src + length;

	while (src < s_end)
	{
		int val = *src++;

		     if (val >  CLIP_THRESHHOLD) val =  CLIP_THRESHHOLD;
		else if (val < -CLIP_THRESHHOLD) val = -CLIP_THRESHHOLD;

		*dest++ = (s16_t) (val >> (16-SAFE_BITS));
	}
}


static void Mix_Mono(mix_channel_t *chan, int *dest, int pairs)
{
	//SYS_ASSERT(pairs > 0);

	sound_t * sfx = (sound_t *) *chan->sfx_ptr;

	const s16_t *src = sfx->samples;

	int *d_pos = dest;
	int *d_end = d_pos + pairs;

	fixed22_t offset = chan->offset;

	while (d_pos < d_end)
	{
		*d_pos++ += src[offset >> 10] * chan->volume_L;

		offset += chan->delta;
	}

	chan->offset = offset;

	//SYS_ASSERT(offset - chan->delta < chan->length);
}


static void Mix_Stereo(mix_channel_t *chan, int *dest, int pairs)
{
	//SYS_ASSERT(pairs > 0);

	sound_t * sfx = (sound_t *) *chan->sfx_ptr;

	const s16_t *src_L = sfx->samples;
	const s16_t *src_R = sfx->samples;
	
	int *d_pos = dest;
	int *d_end = d_pos + pairs * 2;

	fixed22_t offset = chan->offset;

	if (sfx->channels == 2)
		src_R += sfx->length;

	while (d_pos < d_end)
	{
		*d_pos++ += src_L[offset >> 10] * chan->volume_L;
		*d_pos++ += src_R[offset >> 10] * chan->volume_R;

		offset += chan->delta;
	}

	chan->offset = offset;

	//SYS_ASSERT(offset - chan->delta < chan->length);
}


static void Mix_Interleaved(mix_channel_t *chan, int *dest, int pairs)
{
	sound_t * sfx = (sound_t *) *chan->sfx_ptr;

	//SYS_ASSERT(pairs > 0);

	const s16_t *src = sfx->samples;

	int *d_pos = dest;
	int *d_end = d_pos + pairs * 2;

	fixed22_t offset = chan->offset;

	while (d_pos < d_end)
	{
		register fixed22_t pos = (offset >> 9) & ~1;

		*d_pos++ += src[pos  ] * chan->volume_L;
		*d_pos++ += src[pos|1] * chan->volume_R;

		offset += chan->delta;
	}

	chan->offset = offset;

	//SYS_ASSERT(offset - chan->delta < chan->length);
}


#ifdef QUEUE_SUPPORT
static void S_SetupMusicChannel(int q_index);
#endif


static void Mix_OneChannel(mix_channel_t *chan, int pairs)
{
	sound_t * sfx = (sound_t *) *chan->sfx_ptr;

	// cached data gone?
	if (! sfx)
	{
		chan->state = CS_EMPTY;
		chan->sfx_ptr = NULL;
		return;
	}

	if (chan->volume_L == 0 && chan->volume_R == 0)
		return;

	//SYS_ASSERT(chan->offset < chan->length);

	int *dest = mix_buffer;
	
	while (pairs > 0)
	{
		int count = pairs;

		// check if enough sound data is left
		if (chan->offset + count * chan->delta >= chan->length)
		{
			// find minimum number of samples we can play
			double avail = (chan->length - chan->offset + chan->delta - 1) /
				           (double)chan->delta;

			count = (int)floor(avail);

			//SYS_ASSERT(count > 0);
			//SYS_ASSERT(count <= pairs);

			//SYS_ASSERT(chan->offset + count * chan->delta >= chan->length);
		}

		if (snd.channels >= 2 && sfx->interleaved)
			Mix_Interleaved(chan, dest, count);

		else if (snd.channels >= 2)
			Mix_Stereo(chan, dest, count);

		else
			Mix_Mono(chan, dest, count);

		if (chan->offset >= chan->length)
		{
			if (chan == &music_chan)
			{
				// reached end of current queue buffer, get next
				q_bufs[playing_q_index].playing = false;

				playing_q_index = (playing_q_index + 1) % MAX_QUEUE_BUFS;

				if (q_bufs[playing_q_index].playing)
				{
					S_SetupMusicChannel(playing_q_index);
				}
				else
				{
					chan->state = CS_FINISHED;
					// queue_underruns++;
					break;
				}
			}
			else if (chan->loop)
			{
				// we are looping, so clear flag.  The sound needs to
				// be "pumped" (played again) to continue looping.
				chan->loop = false;

				chan->offset = 0;
			}
			else
			{
				chan->state = CS_FINISHED;
				break;
			}
		}

		dest  += count * snd.channels;
		pairs -= count;
	}
}


void S_MixAllChannels(void * udata, byte * stream, int len)
{
	int k;

	if (snd.disabled || len <= 0)
		return;

	int pairs = len / snd.sample_size;

	int samples = pairs * snd.channels;

	// check that we're not getting too much data
	//SYS_ASSERT(pairs <= snd.frag_pairs);

	//SYS_ASSERT(mix_buffer && samples <= mix_buf_len);

	// clear mixer buffer
	memset(mix_buffer, 0, mix_buf_len * sizeof(int));

#if 0  // TESTING.. TESTING..
	mix_buffer[ 0] =  CLIP_THRESHHOLD;
#endif

	// add each channel
	for (k = 0 ; k < NUM_CHANNELS ; k++)
	{
		if (mix_chan[k].state == CS_PLAYING)
		{
			Mix_OneChannel(&mix_chan[k], pairs);
		}
	} 

#ifdef QUEUE_SUPPORT
	if (music_chan.state == CS_PLAYING)
	{
		Mix_OneChannel(&music_chan, pairs);
	}
#endif

	// blit to the SDL stream
	if (snd.bits == 8)
	{
		if (snd.is_signed)
			BlitToS8(mix_buffer, (s8_t *)stream, samples);
		else
			BlitToU8(mix_buffer, (u8_t *)stream, samples);
	}
	else
	{
		if (snd.is_signed)
			BlitToS16(mix_buffer, (s16_t *)stream, samples);
		else
			BlitToU16(mix_buffer, (u16_t *)stream, samples);
	}
}


//----------------------------------------------------------------------------
//  CHANNEL MANAGEMENT
//----------------------------------------------------------------------------

void S_InitChannels(void)
{
	// NOTE: assumes audio is locked!

	int k;

	for (k = 0 ; k < NUM_CHANNELS ; k++)
	{
		mix_chan[k].state = CS_EMPTY;
	}

	// allocate mixer buffer
	mix_buf_len = snd.frag_pairs * snd.channels;

	mix_buffer  = Hunk_AllocName(mix_buf_len * sizeof(int), "blit");

	sfx_start_count = 0;
}


void S_FreeChannels(void)
{
	// NOTE: assumes audio is locked!
	
	int k;

	for (k = 0 ; k < NUM_CHANNELS ; k++)
	{
		S_KillChannel(k);
	}
}


void S_SetupChannel(int k, void ** sfx_ptr, int category)
{
	// NOTE: assumes audio is locked!

	mix_channel_t *chan = &mix_chan[k];

	sound_t * sfx = (sound_t *) *sfx_ptr;


	chan->state = CS_PLAYING;

	chan->sfx_ptr   = sfx_ptr;
	chan->category  = category;
	chan->time      = sfx_start_count++;

	// TODO: entity ref

    // volume computed in S_UpdateSounds() -- just before mixing
    chan->volume_L = 0;
    chan->volume_R = 0;

	/* FIXME */	ComputeVolume(chan);

	chan->offset = 0;
    chan->length = sfx->length << 10;

	ComputeDelta(chan, sfx);

	chan->loop = false;
}


#ifdef QUEUE_SUPPORT
static void S_SetupMusicChannel(int q_index)
{
	volatile queue_sound_t * q_buf = & q_bufs[q_index];

	mix_channel_t *chan = & music_chan;


	chan->state = CS_PLAYING;

	chan->sfx_ptr   = (void **) &q_buf->sfx;
	chan->category  = CAT_UI;
	chan->time      = 0;

    // volume computed in S_UpdateSounds() -- just before mixing
    chan->volume_L = 0;
    chan->volume_R = 0;

	ComputeMusicVolume(chan);

	chan->offset = 0;
    chan->length = q_buf->sfx->length << 10;

	ComputeDelta(chan, q_buf->sfx);

	chan->loop = false;
}
#endif


void S_KillChannel(int k)
{
	// NOTE: assumes audio is locked!

	mix_channel_t *chan = &mix_chan[k];

	if (chan->state != CS_EMPTY)
	{
		chan->state = CS_EMPTY;
		chan->sfx_ptr = NULL;
	}
}


void S_UpdateSounds(vec3_t *listener, float angle)
{
	// NOTE: assume SDL_LockAudio has been called

	int k;

	listen_x = listener ? (*listener)[0] : 0;
	listen_y = listener ? (*listener)[1] : 0;
	listen_z = listener ? (*listener)[2] : 0;

	listen_angle = angle;

	for (k = 0 ; k < NUM_CHANNELS ; k++)
	{
		mix_channel_t *chan = &mix_chan[k];

		if (chan->state == CS_PLAYING)
			ComputeVolume(chan);

		else if (chan->state == CS_FINISHED)
			S_KillChannel(k);
	}

	ComputeMusicVolume(&music_chan);
}


void S_ChangeSoundVolume(void)
{
	/* nothing to do */
}


//----------------------------------------------------------------------------
//  QUEUE SYSTEM
//----------------------------------------------------------------------------

#ifdef QUEUE_SUPPORT


void S_Queue_Init(void)
{
	// we assume the audio stopped / locked

	int i;

	if (snd.disabled)
		return;

	
	free_q_index = 0;

	playing_q_index = 0;


	// allocate all the buffers, and link them into the free list

	for (i = 0 ; i < MAX_QUEUE_BUFS ; i++)
	{
		// create these sound_ts on the hunk (rate is dummy)
		q_bufs[i].sfx = SND_Create (QUEUE_BUF_SAMPLES, 777, 1, NULL);

		q_bufs[i].playing = false;
	}


	music_chan.state    = CS_EMPTY;
	music_chan.sfx_ptr  = NULL;
	music_chan.category = CAT_UI;

	ComputeMusicVolume(&music_chan);
}


void S_Queue_Shutdown(void)
{
	// we assume the audio stopped / locked

	if (snd.disabled)
		return;

	music_chan.state   = CS_EMPTY;
	music_chan.sfx_ptr = NULL;

	free_q_index = 0;

	playing_q_index = 0;
}


sound_t * S_Queue_GetFreeBuffer(void)
{
	sound_t *buf = NULL;

	if (snd.disabled)
		return NULL;

	S_Lock();
	{
		// require the next N buffers are finished
		int future = (free_q_index + 10) % MAX_QUEUE_BUFS;

		if (! q_bufs[future].playing)
		{
			buf = q_bufs[free_q_index].sfx;

			buf->length   = QUEUE_BUF_SAMPLES;
			buf->channels = 1;
		}
	}
	S_Unlock();

	return buf;
}


void S_Queue_AddBuffer(void)
{
	//SYS_ASSERT(! snd.disabled);
	//SYS_ASSERT(buf);

	S_Lock();
	{
		q_bufs[free_q_index].playing = true;

		free_q_index = (free_q_index + 1) % MAX_QUEUE_BUFS;

		if (music_chan.state != CS_PLAYING)
		{
			// either this is the very first buffer to play, or there
			// was an underrun and the queue mixing rover should have
			// caught up with the buffer we just marked to play.
			if (! q_bufs[playing_q_index].playing)
			{
				S_Unlock();
				Sys_Error("Queue system went walkabout into lala land\n");
			}

			S_SetupMusicChannel(playing_q_index);
		}
	}
	S_Unlock();
}


#endif  // QUEUE_SUPPORT

//--- editor settings ---
// vi:ts=4:sw=4:noexpandtab
