/*
===========================================================================
Copyright (C) 1999-2005 Id Software, Inc.
Copyright (C) 2005 Stuart Dalton (badcdev@gmail.com)
Copyright (C) 2005-2006 Joerg Dietrich <dietrich_joerg@gmx.de>
Copyright (C)      2012 Andrew Apted

This file is part of Quake III Arena source code.

Quake III Arena source code is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

Quake III Arena source code is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Quake III Arena source code; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
===========================================================================
*/

#include "main.h"
#include "snd_blit.h"

#include <SDL/SDL_endian.h>


// includes for the OGG codec
#include <errno.h>
#include <vorbis/vorbisfile.h>


// The OGG codec can return the samples in a number of different formats,
// we require the standard signed short format.
#define OGG_SAMPLEWIDTH  2


static OggVorbis_File  vf;

static FILE * s_ogg_file;

static long   s_ogg_beginpos;  // for when the OGG is in a PAK file
static long   s_ogg_filesize;  //

static long   s_ogg_curpos;


//====== callbacks for vorbisfile ======


// fread() replacement
static size_t S_OGG_Callback_read(void *ptr, size_t size, size_t nmemb, void *datasource)
{
	int bytesLeft;

//	int result;


	// check that OGG file is open
	if (! s_ogg_file)
	{
		errno = EBADF; 
		return -1;
	}

	// check if input is valid
	if (! ptr)
	{
		errno = EFAULT; 
		return 0;
	}
	
	if (! (size && nmemb))
	{
		// It's not an error, caller just wants zero bytes!
		errno = 0;
		return 0;
	}

	
	bytesLeft = s_ogg_beginpos + s_ogg_filesize - s_ogg_curpos;

	if (bytesLeft <= 0)  // EOF
	{
		errno = 0;
		return 0;
	}


	s_ogg_curpos += size * nmemb;

	return fread(ptr, size, nmemb, s_ogg_file);
}


// fseek() replacement
static int S_OGG_Callback_seek(void *datasource, ogg_int64_t offset, int whence)
{
	// check that OGG file is open
	if (! s_ogg_file)
	{
		errno = EBADF; 
		return -1;
	}

	switch (whence)
	{
		case SEEK_SET:
		{
			s_ogg_curpos = s_ogg_beginpos + (long) offset;

			return fseek(s_ogg_file, s_ogg_curpos, SEEK_SET);
		}
  
		case SEEK_CUR:
		{
			s_ogg_curpos += (long) offset;

			return fseek(s_ogg_file, s_ogg_curpos, SEEK_SET);
		}
 
		case SEEK_END:
		{
			s_ogg_curpos = s_ogg_beginpos + s_ogg_filesize + (long) offset;

			return fseek(s_ogg_file, s_ogg_curpos, SEEK_SET);
		}
  
		default:
			break;
	}

	// unknown whence, so we return an error
	errno = EINVAL;
	return -1;
}


// fclose() replacement
static int S_OGG_Callback_close(void *datasource)
{
	// we do nothing here and close all things manually in S_OGG_Close()
	return 0;
}


// ftell() replacement
static long S_OGG_Callback_tell(void *datasource)
{
	// check that OGG file is open
	if (! s_ogg_file)
	{
		errno = EBADF; 
		return -1;
	}

	// make sure our own position is up-to-date
	s_ogg_curpos = ftell(s_ogg_file);

	return s_ogg_curpos - s_ogg_beginpos;
}


// the callback structure
const ov_callbacks S_OGG_Callbacks =
{
	&S_OGG_Callback_read,
	&S_OGG_Callback_seek,
	&S_OGG_Callback_close,
	&S_OGG_Callback_tell
};


//----------------------------------------------------------------------


void S_OGG_Init(void)
{
	// TODO: alloc buffers
}


bool_t S_OGG_Open(const char *name)
{
	// check if input is valid
	if (! name)
		return false;

	if (s_ogg_file)
		S_OGG_Close();

	// Open the stream
	if (COM_FOpenFile (name, &s_ogg_file) < 0)
	{
		Con_Printf ("OGG: no such file: %s\n", name);
		return false;
	}

	s_ogg_beginpos = ftell(s_ogg_file);
	s_ogg_filesize = com_filesize;
	s_ogg_curpos   = s_ogg_beginpos;

	// open the codec with our callbacks
	// (passing a dummy thing as datasource, I don't know if NULL is legal)
	if (ov_open_callbacks(s_ogg_file, &vf, NULL, 0, S_OGG_Callbacks) != 0)
	{
		fclose(s_ogg_file);
		s_ogg_file = NULL;

		return false;
	}

	// we only support OGGs with one substream
	if (ov_streams(&vf) != 1)
	{
		Con_Printf("OGG: multiple substreams is not supported\n");

		S_OGG_Close();
		return false;  
	}

#if 0  // NEED THIS ???

	// get the number of sample-frames in the OGG
	numSamples = ov_pcm_total(&vf, 0);

	// fill in the info-structure in the stream
	stream->info.rate = OGGInfo->rate;
	stream->info.width = OGG_SAMPLEWIDTH;
	stream->info.channels = OGGInfo->channels;
	stream->info.samples = numSamples;
	stream->info.size = stream->info.samples * stream->info.channels * stream->info.width;
	stream->info.dataofs = 0;

	// We use stream->pos for the file pointer in the compressed ogg file 
	stream->pos = 0;
	
	// We use the generic pointer in stream for the OGG codec control structure
	stream->ptr = vf;

#endif

	return true;  // OK
}


void S_OGG_Close(void)
{
	// check that OGG file is open
	if (! s_ogg_file)
		return;
	
	// let the OGG codec cleanup its stuff
	ov_clear(&vf);

	fclose(s_ogg_file);
	s_ogg_file = NULL;
}


static int StreamIntoBuffer(sound_t * buf)
{
	int bytesWanted;
	int bytesRead;

	int bitstream;
	int num_samples;

	// big endian machines want their samples in big endian order
	int IsBigEndian = 0;

	if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
		IsBigEndian = 1;

	bytesWanted = buf->length * 2;


	// read a packet of data
	bytesRead = ov_read(&vf, (char *) buf->samples, bytesWanted, IsBigEndian,
						2 /* sample width */, 1 /* signed */, &bitstream);

	// corruption?  skip it
	if (bytesRead == OV_HOLE)
		return 0;
	
	// EOF?  rewind and continue later
	if (bytesRead == 0)
	{
		ov_raw_seek(&vf, 0);
		return 0;
	}

	// error?  gotta kill it
	if (bytesRead < 0)
	{
		Con_Printf("OGG: read error occured (%d)\n", bytesRead);
		return -1;
	}

	num_samples = bytesRead / 2 / buf->channels;

	return num_samples;  // OK
}


// try to enqueue some sound data from the OGG file
void S_OGG_Update(void)
{
	// check that OGG file is open
	if (! s_ogg_file)
		return;

	// fill as many buffers as possible

	for (;;)
	{
		vorbis_info * info;

		sound_t * buf;

		int count;


		buf = S_Queue_GetFreeBuffer();

		// all buffers are in use (good!)
		if (! buf)
			return;


		// determine current rate and channels
		info = ov_info(&vf, -1);

		if (! info)
		{
			Con_Printf("OGG: error occurred (cannot get info)\n");
			S_OGG_Close();
			return;
		}

		if (info->channels < 1 || info->channels > 2)
		{
			Con_Printf("OGG: too many channels (%d)\n", info->channels);
			S_OGG_Close();
			return;
		}

		buf->rate        = info->rate;
		buf->channels    = info->channels;
		buf->interleaved = true;

		count = StreamIntoBuffer(buf);

		if (count > 0)  // OK got some data
		{
			buf->length = count;

			S_Queue_AddBuffer();
			continue;
		}

		if (count < 0)  // error
		{
			S_OGG_Close();
			return;
		}

		// read no data -- file has rewound to beginning.
		// give up for now, and continue on next S_OGG_Update()
		return;
	}
}


//--- editor settings ---
// vi:ts=4:sw=4:noexpandtab
