/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// snd_play.c: picking channel to play on, sound table, etc...


#include "main.h"
#include "snd_blit.h"


// andrewj: the sound table stores sound_ts, which are loaded from WAV files.
//          Each sound_t is in a cache_user_t, hence can become invalidated
//          at any time due to the cache getting full or hunk allocations.
//
//          This sound table has a limited size and it is an error if it
//          becomes full.  (That is easier than managing an LRU cache).
//          It is designed for reasonably fast name lookup, using a hash
//          of the name -- that makes it a hash table I guess.

#define MAX_SOUND_PATH  128

#define SOUND_TABLE_SIZE  1031  // must be a prime number!
#define SOUND_HASH_MASK   1023  // largest power-of-two less than table size


typedef struct cache_sound_s
{
	char  name[MAX_SOUND_PATH];

	cache_user_t  cache;  // the 'data' field points to an sound_t

} cache_sound_t;


static cache_sound_t * sound_table;


int S_FindSound (const char * name)
{
	u32_t hash = COM_StringHash(name);

	int free_pos = -1;

	int step = (hash & SOUND_HASH_MASK) + 1;
	int pos;

	// once pos has reached zero, we know we have tried every position
	// in the sound table (except zero).  This is guaranteed by the
    // fact that SOUND_TABLE_SIZE is a prime number.

	for (pos = step ; pos != 0 ; pos = (pos + step) % SOUND_TABLE_SIZE)
	{
		if (sound_table[pos].name[0])
		{
			if (strcmp(name, sound_table[pos].name) == 0)
				return pos;
		}
		else
		{
			if (free_pos < 0)
				free_pos = pos;
		}
	}

	if (free_pos < 0)
		Sys_Error("Sound table is full\n");


	// store the new sound's filename

	if (strlen(name) > MAX_SOUND_PATH-4)
		Sys_Error ("Sound path too long: %s\n", name);

	strcpy(sound_table[free_pos].name, name);

	sound_table[free_pos].cache.data = NULL;

	return free_pos;
}


void S_CacheSound (int sound_num)
{
	cache_sound_t * cs;

	cs = & sound_table[sound_num];

	// sound data is already in the cache?
	if (cs->cache.data)
	{
		Cache_Check (& cs->cache);

		if (cs->cache.data)
			return;
	}

	// no, hence need to (re-) load the sound from disk
	S_CacheWAV (cs->name, & cs->cache);

	if (! cs->cache.data)
		Sys_Error ("S_CacheSound: failed to load: %s", cs->name);
}


//----------------------------------------------------------------------


static void ReallyStartFX(int sound_num, int category, float vol)
{
	// audio is locked while in here

	cache_sound_t * cs;

	int cat_allowed = (category == CAT_PLAYER || category == CAT_MONSTER) ? 8 : 4;
	int cat_playing = 0;

	int free_k = -1;

	int kill_k    = -1;
	int kill_time = -1;

	int k;


	// count the number of playing channels for this category
	// also find (a) lowest channel that is free
	//           (b) longest playing same-cat channel

	for (k = 0 ; k < NUM_CHANNELS ; k++)
	{
		if (mix_chan[k].state != CS_PLAYING)
		{
			if (free_k < 0)
				free_k = k;
			continue;
		}

		if (mix_chan[k].category != category)
			continue;

		cat_playing++;

		// determine which playing channel to cut off (if we have to)
		// TODO: better score method for positional sounds (prefer closer sounds)
		if (kill_k < 0 || mix_chan[k].time < kill_time)
		{
			kill_k    = k;
			kill_time = mix_chan[k].time;
		}
	}


	if (cat_playing < cat_allowed)
	{
		if (free_k < 0)
		{
			S_Unlock();
			Sys_Error("hogs have invaded the sound system!\n");
		}

		k = free_k;
	}
	else
	{
		// the allowed number of channels for this category are all in use.
		// need to kick one out.
		k = kill_k;
	}

	if (mix_chan[k].state != CS_EMPTY)
		S_KillChannel(k);

	// TODO: handle 'vol'

	cs = & sound_table[sound_num];

	S_SetupChannel(k, & cs->cache.data, category);
}


void S_StartFX(int sound_num, int category, float vol)
{
	if (snd.disabled)
		return;

	// ignore very far away sounds
#if 0
	if (category >= CHAN_Monster)
	{
		float dist = P_ApproxDistance(listen_x - pos->x, listen_y - pos->y, listen_z - pos->z);

		if (dist > max_distance)
			return;
	}
#endif

	S_CacheSound(sound_num);

	S_Lock();
	{
		ReallyStartFX(sound_num, category, vol);
	}
	S_Unlock();
}


void S_StopAllSounds(bool_t clear)
{
	// FIXME
}


void S_InitSoundTable(void)
{
	// allocate the sound table

	sound_table = Hunk_AllocName (sizeof(cache_sound_t) * SOUND_TABLE_SIZE, "zndtab");
}


// Called once each time through the main loop
void S_Update(vec3_t origin, vec3_t forward, vec3_t right, vec3_t up)
{
	// FIXME

	S_OGG_Update();
}


void S_ExtraUpdate(void)
{
}


//--- editor settings ---
// vi:ts=4:sw=4:noexpandtab
