/*
===========================================================================
Copyright (C) 1999-2005 Id Software, Inc.
Copyright (C) 2005 Stuart Dalton (badcdev@gmail.com)
Copyright (C) 2012 Andrew Apted

This file is part of Quake III Arena source code.

Quake III Arena source code is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

Quake III Arena source code is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Quake III Arena source code; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
===========================================================================
*/

#include "main.h"


static short GetLittleShort(const byte *pos)
{
	return (short)((pos[1] << 8) | pos[0]);
}


static int GetLittleLong(const byte *pos)
{
	return (int)((pos[3] << 24) | (pos[2] << 16) | (pos[1] << 8) | pos[0]);
}


static int S_ReadChunkInfo(const byte ** pos, char *name)
{
	int len;

	memcpy(name, *pos, 4);  *pos += 4;

	name[4] = 0;


	len = GetLittleLong(*pos);  *pos += 4;

	if (len < 0)
	{
//??	Com_Printf( S_COLOR_YELLOW "WARNING: Negative chunk length\n" );
		return -1;
	}

	return len;
}


// returns the length of the data in the chunk, or -1 if not found
static int S_FindRIFFChunk(const byte ** pos, const byte * p_end, const char *chunk)
{
	char	name[5];
	int		len;

	for ( ; ; )
	{
		if (*pos + 8 > p_end)
			break;

		len = S_ReadChunkInfo(pos, name);

		if (len < 0)
			break;
	
		// If this is the right chunk, return
		if (Q_strncmp(name, chunk, 4) == 0)
			return len;

		// pad to word boundary
		if (len & 1)
			len++;

		// Not the right chunk - skip it
		*pos += len;
	}

	return -1;  // NOT FOUND
}


typedef struct
{
	int format;
	int channels;
	int rate;
	int bits;
	int width;
	int size;
	int length;

} wav_info_t;


static bool_t S_ReadRIFFHeader(const byte ** pos, const byte * p_end, wav_info_t *info)
{
	int fmt_len = 0;

	// skip the file header
	*pos += 12;

	// Scan for the format chunk
	fmt_len = S_FindRIFFChunk(pos, p_end, "fmt ");

	if (fmt_len < 0)
	{
		Con_Printf ("WAV: Couldn't find \"fmt\" chunk\n", info->channels);
		return false;
	}

	// Save the parameters
	info->format = GetLittleShort(*pos);    *pos += 2;
	info->channels = GetLittleShort(*pos);  *pos += 2;
	info->rate = GetLittleLong(*pos);       *pos += 10;
	info->bits = GetLittleShort(*pos);      *pos += 2;

	// !!!! FIXME: validate 'format', 'rate' and 'bits'

	if (info->channels < 1 || info->channels > 2)
	{
		Con_Printf ("WAV: %d channels not supported\n", info->channels);
		return false;
	}

	if (info->bits < 8)
	{
		Con_Printf ("WAV: Less than 8 bit sound is not supported\n");
		return false;
	}

	info->width = info->bits / 8;

	// Skip the rest of the format chunk if required
	if (fmt_len > 16)
	{
		fmt_len -= 16;

		*pos += fmt_len;
	}

	// Scan for the data chunk
	info->size = S_FindRIFFChunk(pos, p_end, "data");

	if (info->size < 0)
	{
		Con_Printf ("WAV: Couldn't find \"data\" chunk\n");
		return false;
	}

	info->length = (info->size / info->width) / info->channels;

	return true;  // OK!
}


static bool_t S_ReadRIFFData_8x1(const byte * pos, const byte * p_end, sound_t *sfx)
{
	short * L = sfx->samples;
	short * L_end = L + sfx->length;

	while (L < L_end)
	{
		*L++ = (short)((*pos++ ^ 0x80) << 8);
	}

	return true;
}


static bool_t S_ReadRIFFData_8x2(const byte * pos, const byte * p_end, sound_t *sfx)
{
	short * L = sfx->samples;
	short * L_end = L + sfx->length;
	short * R = L_end;

	while (L < L_end)
	{
		*L++ = (short)((*pos++ ^ 0x80) << 8);
		*R++ = (short)((*pos++ ^ 0x80) << 8);
	}

	return true;
}


static bool_t S_ReadRIFFData_16x1(const byte * pos, const byte * p_end, sound_t *sfx)
{
	short * L = sfx->samples;
	short * L_end = L + sfx->length;

	while (L < L_end)
	{
		*L++ = GetLittleShort(pos);  pos += 2;
	}

	return true;
}


static bool_t S_ReadRIFFData_16x2(const byte * pos, const byte * p_end, sound_t *sfx)
{
	short * L = sfx->samples;
	short * L_end = L + sfx->length;
	short * R = L_end;

	while (L < L_end)
	{
		*L++ = GetLittleShort(pos);  pos += 2;
		*R++ = GetLittleShort(pos);  pos += 2;
	}

	return true;
}


sound_t * S_CacheWAV(const char * name, struct cache_user_s * cu)
{
	byte * data;

	const byte * pos;
	const byte * p_end;

	wav_info_t info;

	sound_t * sfx;


	data = COM_LoadTempFile(name);

	if (! data)
	{
		Con_Printf ("Missing sound file: %s\n", name);
		return NULL;
	}

	
	// Read the RIFF header

	if (com_filesize < 64)
	{
		Con_Printf ("Bad WAV file: %s (too small)\n", name);
		return NULL;
	}

	pos   = data;
	p_end = data + com_filesize;


	if (! S_ReadRIFFHeader(&pos, p_end, &info))
	{
		Con_Printf ("Bad WAV file: %s (bad format)\n", name);
		return NULL;
	}


	// allocate the sound_t + sample data following it

	sfx = SND_Create(info.length, info.rate, info.channels, cu);

	if (! sfx)
		return NULL;


	// Read data....

	if (info.bits == 16)
	{
		if (info.channels == 2)
			S_ReadRIFFData_16x2(pos, p_end, sfx);
		else
			S_ReadRIFFData_16x1(pos, p_end, sfx);
	}
	else  // bits == 8
	{
		if (info.channels == 2)
			S_ReadRIFFData_8x2(pos, p_end, sfx);
		else
			S_ReadRIFFData_8x1(pos, p_end, sfx);
	}

	return sfx;
}


//--- editor settings ---
// vi:ts=4:sw=4:noexpandtab
