/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c -- this is the only file outside the refresh that touches the
// vid buffer

#include "main.h"

static image_t		* draw_chars;
static image_t		* draw_bigchars;

static image_t		* draw_conback;


#define FONT_NAME	"font22.tga"
#define FONT_DX		12
#define FONT_W		12
#define FONT_H		22

#define BIGFONT_NAME	"spikey_font6.tga"
#define BIGFONT_DX		16
#define BIGFONT_W		25
#define BIGFONT_H		25


//=============================================================================
/* Support Routines */


// andrewj: the image table stores image_ts, which are loaded from TGA files.
//          Each image_t is in a cache_user_t, hence can become invalidated
//          at any time due to the cache getting full or hunk allocations.
//
//          This image table has a limited size and it is an error if it
//          becomes full.  (That is easier than managing an LRU cache).
//          It is designed for reasonably fast name lookup, using a hash
//          of the name -- that makes it a hash table I guess.

#define MAX_IMAGE_PATH  128

#define IMAGE_TABLE_SIZE  1031  // must be a prime number!
#define IMAGE_HASH_MASK   1023  // largest power-of-two less than table size


typedef struct cache_image_s
{
	char  name[MAX_IMAGE_PATH];

	cache_user_t  cache;  // the 'data' field points to an image_t

} cache_image_t;


static cache_image_t * image_table;


static int V_FindCacheImage (const char * name)
{
	u32_t hash = COM_StringHash(name);

	int free_pos = -1;

	int step = (hash & IMAGE_HASH_MASK) + 1;
	int pos;

	// once pos has reached zero, we know we have tried every position
	// in the image table (except zero).  This is guaranteed by the
    // fact that IMAGE_TABLE_SIZE is a prime number.

	for (pos = step ; pos != 0 ; pos = (pos + step) % IMAGE_TABLE_SIZE)
	{
		if (image_table[pos].name[0])
		{
			if (strcmp(name, image_table[pos].name) == 0)
				return pos;
		}
		else
		{
			if (free_pos < 0)
				free_pos = pos;
		}
	}

	if (free_pos < 0)
		Sys_Error("Image table is full\n");

	return free_pos;
}


image_t * V_CacheImage (const char *name)
{
	int  pos;

	image_t * img;


	pos = V_FindCacheImage (name);

	if (! image_table[pos].name[0])
	{
		// adding a new image
		if (strlen(name) > MAX_IMAGE_PATH-4)
			Sys_Error ("Image path too long: %s\n", name);

		strcpy(image_table[pos].name, name);

		image_table[pos].cache.data = NULL;
	}
	else if (image_table[pos].cache.data)
	{
		img = (image_t *) Cache_Check(& image_table[pos].cache);

		if (img)
			return img;
	}

	// need to (re-) load the image from disk

	img = V_CacheTGA (name, & image_table[pos].cache);

	if (! img)
		Sys_Error ("V_CacheImage: failed to load: %s", name);
	
	return img;
}


// part coordinates include x1 but not x2, include y1 but not y2
void V_DrawImagePart (int sx, int sy, image_t * img,
                      int x1, int y1, int x2, int y2)
{
	int x, y;

	// clip to screen
	if (sx < 0)
	{
		x1 += -sx;  sx = 0;
	}
	if (sx + x2 - x1 > vid.width)
	{
		x2 = vid.width - sx + x1;
	}

	if (sy < 0)
	{
		y1 += -sy;  sy = 0;
	}
	if (sy + y2 - y1 > vid.height)
	{
		y2 = vid.height - sy + y1;
	}


	// totally clipped?
	if (x2 < x1 || y2 < y1)
		return;


	for (y = y1 ; y < y2 ; y++)
	for (x = x1 ; x < x2 ; x++)
	{
		rgbcol_t pix = img->pixels[y * img->width + x];

		if (RGB_A(pix) == 0)
			continue;

		*((rgbcol_t *)(vid.buffer + (sy+y-y1) * vid.rowbytes) + (sx+x-x1)) = pix;
	}
}


void V_DrawCacheImage (int x, int y, const char *name)
{
	image_t * img = V_CacheImage(name);

	V_DrawImagePart(x, y, img, 0, 0, img->width, img->height);
}


void V_DrawCacheTile (int x, int y, const char *name, int tw, int th, int tx, int ty)
{
	image_t * img = V_CacheImage(name);

	int px, py, pw, ph;

#if 0
	if (isdigit(part[0]))
		px = part[0] - '0';
	else
		px = (part[0] | 32) - 'a';

	if (isdigit(part[1]))
		py = part[1] - '0';
	else
		py = (part[1] | 32) - 'a';
	
	pw = ph = 1;

	if (part[2])
	{
		if (isdigit(part[2]))
			pw = part[2] - '0';
		else
			pw = (part[2] | 32) - 'a';

		if (isdigit(part[3]))
			ph = part[3] - '0';
		else
			ph = (part[3] | 32) - 'a';
	}
#else
	px = tx;  pw = 1;
	py = ty;  ph = 1;
#endif

	if ((px + pw) * tw > img->width ||
	    (py + ph) * th > img->height)
	{
		Sys_Error("draw_tile: bad part %dx%d requested for: %s\n", tx, ty, name); 
	}

	V_DrawImagePart(x, y, img, px*tw, py*th, (px+pw)*tw, (py+ph)*th);
}


/*
===============
Draw_Init
===============
*/
void Draw_Init (void)
{
	// allocate the image table

	image_table = Hunk_AllocName (sizeof(cache_image_t) * IMAGE_TABLE_SIZE, "imgtab");

	// these are always needed : load on hunk

	draw_chars = V_CacheTGA (FONT_NAME, NULL);
	if (! draw_chars)
		Sys_Error("Unable to load font!\n");

	draw_bigchars = V_CacheTGA (BIGFONT_NAME, NULL);
	if (! draw_bigchars)
		Sys_Error("Unable to load big font!\n");

#if 0
	draw_conback = V_CacheTGA ("conback.tga", NULL);
	if (! draw_conback)
		Sys_Error("Unable to load conback!\n");
#endif

}



/*
================
Draw_Character

Draws one 8*8 graphics character with 0 being transparent.
It can be clipped to the top of the screen to allow the console to be
smoothly scrolled off.
================
*/
void Draw_Character (int x, int y, int num)
{
	byte			*dest;
	rgbcol_t		*source;
	int				drawline;	
	int				row, col;

	rgbcol_t   color;

	color = (num & 0x80) ? 0xFFFF00 : 0xFFFFFF;  // yellow / white

	num &= 0x7f;
	
	if (y <= -FONT_H)
		return;			// totally off screen

	if (y > vid.height - FONT_H || x < 0 || x > vid.width - FONT_W)
		return;

#ifdef PARANOID
	if (num < 0 || num > 255)
		Sys_Error ("Con_DrawCharacter: char %i", num);
#endif

	row = num >> 4;
	col = num & 15;

	source = draw_chars->pixels + ((row * FONT_H * 16) + col) * FONT_W;

	if (y < 0)
	{	// clipped
		drawline = FONT_H + y;
		source -= y * 16 * FONT_W;
		y = 0;
	}
	else
		drawline = FONT_H;


	{
		rgbcol_t *d2 = (rgbcol_t *)(vid.buffer + y*vid.rowbytes) + x;
	
		while (drawline--)
		{
			int dx;

			for (dx = 0 ; dx < FONT_W ; dx++)
				if (source[dx] & 0x8000)
					d2[dx] = color;

			source += 16 * FONT_W;

			d2 += vid.y_step;
		}
	}
}


static void Draw_BigCharacter (int x, int y, int num, int color)
{
	byte			*dest;
	rgbcol_t		*source;
	int				drawline;	
	int				row, col;

	num &= 0x7f;
	
	if (y < -BIGFONT_H || y > vid.height || x < -BIGFONT_H || x > vid.width)
		return;

	if (num == ':')
		num = 26;
	else if (num == '?')
		num = 27;
	else if (isalpha(num))
		num = toupper(num) - 'A';
	else
		return;  // not in font

	col = (num % 14) * BIGFONT_W;
	row = (num / 14 + (color & 3) * 2) * BIGFONT_H;

	V_DrawImagePart(x, y, draw_bigchars, col, row, col+BIGFONT_W-1, row+BIGFONT_H-1);
}


/*
================
Draw_String
================
*/
void Draw_String (int x, int y, const char *str)
{
	while (*str)
	{
		Draw_Character (x, y, *str++);

		x += FONT_DX;
	}
}

void Draw_BigString (int x, int y, const char *str, int color)
{
	while (*str)
	{
		Draw_BigCharacter (x, y, *str++, color);

		x += BIGFONT_DX;
	}
}


void Draw_ImageT (int x, int y, image_t * img)
{
	int dx, dy;

	for (dy = 0 ; dy < img->height ; dy++)
	for (dx = 0 ; dx < img->width  ; dx++)
	{
		rgbcol_t pix = img->pixels[dy * img->width + dx];

		if (RGB_A(pix) == 0)
			continue;

//		vid.buffer[(y+dy) * vid.rowbytes + x+dx] = RGB_G(pix) / 17;
		*((rgbcol_t *)(vid.buffer + dy * vid.rowbytes) + dx) = pix;
	}
}


#if 0
/*
=============
Draw_Pic
=============
*/
void Draw_Pic (int x, int y, qpic_t *pic)
{
	byte			*source;
	rgbcol_t        *dest;

	int  dx, dy;

	if ((x < 0) ||
		(x + pic->width > vid.width) ||
		(y < 0) ||
		(y + pic->height > vid.height))
	{
		Sys_Error ("Draw_Pic: bad coordinates");
	}

	source = pic->data;


	for (dy = 0 ; dy < pic->height ; dy++, source += pic->width)
	for (dx = 0 ; dx < pic->width  ; dx++)
	{
		byte pix = source[dx];

		if (pix == TRANSPARENT_COLOR)
			continue;

		dest = (rgbcol_t *)(vid.buffer + (y+dy) * vid.rowbytes) + (x+dx);

		if (128 <= pix && pix <= 223)
			pix ^= 15;

		pix &= 15;

		*dest = pix * 0x101010;
	}
}


/*
=============
Draw_TransPic
=============
*/
void Draw_TransPic (int x, int y, qpic_t *pic)
{
	Draw_Pic(x, y, pic);
}


/*
=============
Draw_TransPicTranslate
=============
*/
void Draw_TransPicTranslate (int x, int y, qpic_t *pic, byte *translation)
{
	Draw_Pic(x, y, pic);
}
#endif


void Draw_CharToConback (int num, byte *dest)
{
#if 0
	int		row, col;
	byte	*source;
	int		drawline;
	int		x;

	row = num>>4;
	col = num&15;
	source = draw_chars + (row<<10) + (col<<3);

	drawline = 8;

	while (drawline--)
	{
		for (x=0 ; x<8 ; x++)
			if (source[x])
				dest[x] = 0x60 + source[x];
		source += 128;
		dest += 320;
	}
#endif
}

/*
================
Draw_ConsoleBackground

================
*/
void Draw_ConsoleBackground (int lines)
{
	int				x, y, sy;
	byte			*dest;
	rgbcol_t   *src;
	byte   pix;

	dest = vid.buffer;

	for (y = 0 ; y < lines ; y++, dest += vid.rowbytes)
	{
#if 0
		sy = (draw_conback->height - lines + y)/2;
		if (sy < 0)
		{
			sy = -sy;

			if (sy >= draw_conback->height)
				sy = draw_conback->height * 2 - sy;

			if (sy < 0) break;
		}

		pix = 34; // 175 - ((lines - y) / 4) & 3;
#endif

#if 1
		if (y >= lines-2)
			pix = 96;
		else
			pix = 32;

		memset(dest, pix, vid.rowbytes);

#else
		for (x = 0 ; x < vid.width ; x++)
		{
			int sx = (x/2) % draw_conback->width;

			rgbcol_t pix = draw_conback->pixels[sy * draw_conback->width + sx];

			// darken
			pix = (pix & 0x00FEFEFE) >> 1;

			rgbcol_t *d2 = (rgbcol_t *)dest;

			d2[x] = pix;
		}
#endif
	}
}


/*
=============
Draw_TileClear

This repeats a 64*64 tile graphic to fill the screen around a sized down
refresh window.
=============
*/
void Draw_TileClear (int x, int y, int w, int h)
{
	// Fixme ??
}


/*
=============
Draw_Fill

Fills a box of pixels with a single color
=============
*/
void Draw_Fill (int x, int y, int w, int h, int c)
{
	byte			*dest;
	unsigned short	*pusdest;
	unsigned		uc;
	int				u, v;

	if (r_pixbytes == 1)
	{
		dest = vid.buffer + y*vid.rowbytes + x;
		for (v=0 ; v<h ; v++, dest += vid.rowbytes)
			for (u=0 ; u<w ; u++)
				dest[u] = c;
	}
	else
	{
		uc = d_8to16table[c];

		pusdest = (unsigned short *)vid.buffer + y * (vid.rowbytes >> 1) + x;
		for (v=0 ; v<h ; v++, pusdest += (vid.rowbytes >> 1))
			for (u=0 ; u<w ; u++)
				pusdest[u] = uc;
	}
}
//=============================================================================

/*
================
Draw_FadeScreen

================
*/
void Draw_FadeScreen (void)
{
	int			x,y;
	byte		*pbuf;

	VID_UnlockBuffer ();
	S_ExtraUpdate ();
	VID_LockBuffer ();

	for (y=0 ; y<vid.height ; y++)
	{
		int	t;

		pbuf = (byte *)(vid.buffer + vid.rowbytes*y);
		t = (y & 1) << 1;

		for (x=0 ; x<vid.width ; x++)
		{
			if ((x & 3) != t)
				pbuf[x] = 0;
		}
	}

	VID_UnlockBuffer ();
	S_ExtraUpdate ();
	VID_LockBuffer ();
}


//--- editor settings ---
// vi:ts=4:sw=4:noexpandtab
