//------------------------------------------------------------------------
//  TGA (Targa) LOADING
//------------------------------------------------------------------------
//
//  Copyright (c) 2006-2012  Andrew Apted
//  Copyright (C) 1997-2006  Sam Lantinga
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//------------------------------------------------------------------------
//
//  Based on IMG_tga.c in the SDL_image library, by Sam Lantinga.
//
//------------------------------------------------------------------------

#include "main.h"


// -AJA- according to the documentation I see, colormaps in TGA files
//       are BGR(A) format (consistent with non-cmap files).
//
//       However it doesn't work with my test case (saved from the GIMP).
//       Needs more testing with other programs.
#define SWAP_FOR_INDEXED  0


typedef struct
{
    byte info_len;        /* length of info field */
    byte has_cmap;        /* 1 if image has colormap, 0 otherwise */
    byte type;

    byte cmap_start[2];   /* index of first colormap entry */
    byte cmap_len[2];     /* number of entries in colormap */
    byte cmap_bits;       /* bits per colormap entry */

    byte y_origin[2];     /* image origin */
    byte x_origin[2];
    byte width[2];        /* image size */
    byte height[2];

    byte pixel_bits;      /* bits/pixel */
    byte flags;
}
tga_header_t;

#define GET_U16_FIELD(hdvar, field)  (u16_t)  \
		((hdvar).field[0] + ((hdvar).field[1] << 8))

#define GET_S16_FIELD(hdvar, field)  (s16_t)  \
		((hdvar).field[0] + ((hdvar).field[1] << 8))

typedef enum
{
	TGA_TYPE_INDEXED = 1,
    TGA_TYPE_RGB = 2,
    TGA_TYPE_BW = 3,

    TGA_TYPE_RLE_INDEXED = 9,
    TGA_TYPE_RLE_RGB = 10,
    TGA_TYPE_RLE_BW = 11
}
tga_type_e;

typedef enum
{
	TGA_FLAG_TOP_DOWN = 0x20,
}
tga_flag_e;


// current decode information

static tga_header_t header;

static byte * decode_pos;
static byte * decode_end;  // end position of data

static int decode_x;
static int decode_y;
static int decode_bpp;


static bool_t ReadPixels(image_t * img, int count)
{
	int k;
	int w;

	while (count > 0)
	{
		if (decode_y >= img->height || decode_pos >= decode_end)
		{
//			Con_Printf("V_CacheTGA: uncompressed too much!\n");
			return false;
		}

		w = img->width - decode_x;

		if (w > count)
			w = count;
		
//!!	SYS_ASSERT(w > 0);
//!!	SYS_ASSERT(x + w <= img->width);

		rgbcol_t * dest = img->pixels + (decode_y * img->width) + decode_x;

		for (k = 0 ; k < w ; k++)
		{
			if (decode_bpp == 3)
			{
				*dest++ = MAKE_RGB(decode_pos[2], decode_pos[1], decode_pos[0]);
			}
			else
			{
				*dest++ = MAKE_RGBA(decode_pos[2], decode_pos[1], decode_pos[0], decode_pos[3]);
			}

			decode_pos += decode_bpp;
		}

		decode_x += w;
		count    -= w;

		if (decode_x >= img->width)
		{
			decode_x  = 0;
			decode_y += 1;
		}
	}

	return true;
}


static bool_t DuplicatePixels(image_t * img, int count, rgbcol_t src)
{
	int w;
	int k;

	while (count > 0)
	{
		if (decode_y >= img->height)
		{
//			Con_Printf("V_CacheTGA: uncompressed too much!\n");
			return false;
		}

		w = img->width - decode_x;

		if (w > count)
			w = count;
		
//!!	SYS_ASSERT(w > 0);
//!!	SYS_ASSERT(x + w <= img->width);

		rgbcol_t * dest = img->pixels + (decode_y * img->width) + decode_x;

		for (k = 0 ; k < w ; k++)
		{
			*dest++ = src;
		}

		decode_x += w;
		count    -= w;

		if (decode_x >= img->width)
		{
			decode_x  = 0;
			decode_y += 1;
		}
	}

	return true;
}


static bool_t DecodeRGB(image_t * img)
{
	return ReadPixels(img, img->width * img->height);
}


static bool_t DecodeRLE_RGB(image_t * img)
{
	while (decode_y < img->height && decode_pos < decode_end)
	{
		/* read the next "packet" */

		byte count = *decode_pos++;

		if ((count & 0x80) == 0) // raw packet
		{
			if (! ReadPixels(img, count + 1))
				return false;
		}
		else // run-length packet
		{
			int save_x = decode_x;
			int save_y = decode_y;

			if (! ReadPixels(img, 1))
				return false;

			rgbcol_t src = img->pixels[save_y * img->width + save_x];

			if (! DuplicatePixels(img, count & 0x7F, src))
				return false;
		}
	}

	return true;
}


// TODO : support INDEXED modes

#if 0
static byte * ReadPalette(file_c *f, tga_header_t &header)
{
	int i;
	int num_colors;

	rgbcol_t palette[256];


	num_colors = header.cmap_len[1] * 256 + header.cmap_len[0];

	if (num_colors <= 0 || num_colors > 256)
		Sys_Error("TGA: illegal number of colors in image: %d\n", num_colors);

	if (header.pixel_bits != 8)
		Sys_Error("TGA: %d bit color-mapped pixels not supported.\n", (int)header.pixel_bits);


	memset(palette, 0, sizeof(palette));

	int bpp = (header.cmap_bits == 32) ? 4 : 3;

	for (i = 0 ; i < num_colors ; i++)
	{
		byte color[4];

		if (bpp != (int)f->Read(&palette[i*4], bpp))
		{
			I_Debugf("V_CacheTGA: error reading palette.\n");
			return NULL;
		}
	}

#if (SWAP_FOR_INDEXED)
	SwapPixels(palette, bpp, num_colors);
#endif

	return palette;
}


static bool_t ReadPixels_ColMap(image_t *img, int& x, int& y,
						      int total, const byte *palette)
{
	byte buffer[256];
		
	while (total > 0)
	{
		if (y >= img->height)
		{
			I_Debugf("V_CacheTGA: uncompressed too much!\n");
			return false;
		}

		int w = img->width - x;

		if (w > total) w = total;
		if (w > 256)   w = 256;

//!!	SYS_ASSERT(w > 0);
//!!	SYS_ASSERT(x + w <= img->width);

		byte *dest = img->PixelAt(x, y);

		if (w != (int)f->Read(buffer, w))
			return false;

		// FIXME: make sure color indices are in range!

		for (int p=0; p < w; p++, dest += img->bpp)
			memcpy(dest, palette + (buffer[p] * 4), img->bpp);

		x     += w;
		total -= w;

		if (x >= img->width)
		{
			x  = 0;
			y += 1;
		}
	}

	return true;
}


static bool_t DecodeIndexed(image_t * img)
{
	rgbcol_t *palette = ReadPalette(data, header);

	if (! palette)
		return false;

	return ReadPixels_ColMap(img, f, x, y, img->width * img->height, palette);
}


static bool_t DecodeRLE_Indexed(image_t * img)
{
	byte *palette = ReadPalette(f, header);
	if (! palette)
		return false;

	while (y < img->height)
	{
		/* read the next "packet" */

		byte count;

		if (1 != f->Read(&count, 1))
			return false;

		if ((count & 0x80) == 0) // raw packet
		{
			if (! ReadPixels_ColMap(img, f, x, y, count + 1, palette))
				return false;
		}
		else // run-length packet
		{
			const byte *src = img->PixelAt(x, y);

			if (! ReadPixels_ColMap(img, f, x, y, 1, palette))
				return false;

			if (! DuplicatePixels(img, x, y, count & 0x7F, src))
				return false;
		}
	}

	return true;
}

#endif  // end of non-working indexed support


image_t * V_CacheTGA(const char * name, struct cache_user_s * cu)
{
	byte * data;

	int width, height;
	bool_t result;

	image_t * img;


	data = COM_LoadTempFile(name);

	if (! data)
	{
		Con_Printf ("Missing image file: %s\n", name);
		return NULL;
	}

	
	// verify the TGA header is valid

	if (com_filesize < sizeof(header))
	{
		Con_Printf ("Bad TGA file: %s (too small)\n", name);
		return NULL;
	}

	memcpy(&header, data, sizeof(header));


	if (header.type != TGA_TYPE_INDEXED &&
        header.type != TGA_TYPE_RGB &&
	    header.type != TGA_TYPE_RLE_INDEXED &&
		header.type != TGA_TYPE_RLE_RGB)
	{
		Con_Printf ("Bad TGA file: %s (unknown type %d)\n", name, header.type);
		return NULL;
	}

	if (header.type == TGA_TYPE_INDEXED ||
		header.type == TGA_TYPE_RLE_INDEXED)
	{
		// if (header.cmap_bits < 24)
		//   return NULL;

		Con_Printf ("Bad TGA file: %s (colormapped not supported)\n", name);
		return NULL;
	}
	else
	{
		if (header.pixel_bits < 24)
		{
			Con_Printf ("Bad TGA file: %s (16 bpp not supported)\n", name);
			return NULL;
		}
	}

	width  = GET_U16_FIELD(header, width);
	height = GET_U16_FIELD(header, height);

	if (width  < 1 || width  > 2048 || 
		height < 1 || height > 2048)
	{
		Con_Printf ("Bad TGA file: %s (weird size: %dx%d)\n", width, height);
		return NULL;
	}


	if (header.type == TGA_TYPE_INDEXED || header.type == TGA_TYPE_RLE_INDEXED)
		decode_bpp = (header.cmap_bits == 32) ? 4 : 3;
	else
		decode_bpp = (header.pixel_bits == 32) ? 4 : 3;


	// move pointer to pixel data, skip any info field
	decode_end = data + com_filesize;

	decode_pos = data + sizeof(header);

	if (header.info_len > 0)
	{
		decode_pos += header.info_len;
	}


#if 0 // DEBUG
	fprintf(stderr, "V_CacheTGA : %s\n", name);
	fprintf(stderr, "  width  = %d\n", width);
	fprintf(stderr, "  height = %d\n", height);
	fprintf(stderr, "  bpp    = %d\n", decode_bpp);
#endif


	img = IMG_Create(width, height, cu);


	/* read image pixels */

	decode_x = 0;
	decode_y = 0;

	switch (header.type)
	{
		case TGA_TYPE_RGB:
			result = DecodeRGB(img);
			break;

		case TGA_TYPE_RLE_RGB:
			result = DecodeRLE_RGB(img);
			break;

#if 0
		case TGA_TYPE_INDEXED:
			result = DecodeIndexed(img);
			break;

		case TGA_TYPE_RLE_INDEXED:
			result = DecodeRLE_Indexed(img);
			break;
#endif

		default:
			result = false;
			break;
	}

	if (! result)
		Con_Printf("WARNING: failure decoding TGA image: %s\n", name);

	if (! (header.flags & TGA_FLAG_TOP_DOWN))
		IMG_Invert(img);

	return img;
}


//--- editor settings ---
// vi:ts=4:sw=4:noexpandtab
